#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2010 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Diagnostics;
using System.Threading;
using HeuristicLab.Common;
using HeuristicLab.Hive.Contracts;
using HeuristicLab.Hive.Contracts.BusinessObjects;
using HeuristicLab.Hive.Slave.Common;
using HeuristicLab.Hive.Slave.Communication;
using HeuristicLab.Hive.Slave.Communication.SlaveFacade;
using HeuristicLab.Hive.Slave.Core.ConfigurationManager;

namespace HeuristicLab.Hive.Slave.Core {
  /// <summary>
  /// Heartbeat class. It sends every x ms a heartbeat to the server and receives a Message
  /// </summary>
  public class HeartbeatManager {
    private static object locker = new object();
    public TimeSpan Interval { get; set; }
    private Thread heartBeatThread;
    private AutoResetEvent waitHandle;

    public HeartbeatManager() {
      Interval = new TimeSpan(0, 0, 10);
    }

    public HeartbeatManager(TimeSpan interval) {
      Interval = interval;
    }

    private WcfService wcfService;

    private bool abortThreadPending;

    ReaderWriterLockSlim heartBeatThreadIsSleepingLock = new ReaderWriterLockSlim();
    private bool heartBeatThreadIsSleeping = false;

    /// <summary>
    /// Starts the Heartbeat signal.
    /// </summary>
    public void StartHeartbeat() {
      this.waitHandle = new AutoResetEvent(true);
      wcfService = WcfService.Instance;
      wcfService.ProcessHeartBeatCompleted += new EventHandler<ProcessHeartBeatCompletedEventArgs>(wcfService_ProcessHeartBeatCompleted);
      abortThreadPending = false;
      heartBeatThread = new Thread(RunHeartBeatThread);
      heartBeatThread.Start();
    }

    /// <summary>
    /// Stop the heartbeat
    /// </summary>
    public void StopHeartBeat() {
      abortThreadPending = true;
      waitHandle.Set();
      heartBeatThread.Join();
    }
    
    /// <summary>
    /// use this method to singalize there is work to do (to avoid the waiting period if its clear that actions are required)
    /// </summary>
    public void AwakeHeartBeatThread() {
      waitHandle.Set();
    }

    private void RunHeartBeatThread() {
      while (!abortThreadPending) {
        try {
          lock (locker) {
            if (wcfService.ConnState != NetworkEnum.WcfConnState.Connected) {
              wcfService.Connect(ConfigManager.Instance.GetClientInfo()); // Login happens automatically upon successufl connection
            }
            if(wcfService.ConnState == NetworkEnum.WcfConnState.Connected) {
              SlaveDto info = ConfigManager.Instance.GetClientInfo();

              HeartBeatData heartBeatData = new HeartBeatData {
                SlaveId = info.Id,
                FreeCores = info.NrOfCores.HasValue ? info.NrOfCores.Value - ConfigManager.Instance.GetUsedCores() : 0,
                FreeMemory = GetFreeMemory(),
                JobProgress = ConfigManager.Instance.GetExecutionTimeOfRunningJobs(),
                IsAllowedToCalculate = UptimeManager.Instance.IsAllowedToCalculate() && UptimeManager.Instance.CalendarAvailable
              };

              if (!heartBeatData.IsAllowedToCalculate) {
                // stop all running jobs and send snapshots to server
                MessageQueue.GetInstance().AddMessage(MessageContainer.MessageType.UptimeLimitDisconnect);
              }

              Logger.Debug("Sending Heartbeat: " + heartBeatData);
              wcfService.ProcessHeartBeatSync(heartBeatData);
            }
          } // lock
        }
        catch (Exception e) {
          Logger.Error("Heartbeat Thread failed badly: " + e.Message);
          OnExceptionOccured(e);
        }
        waitHandle.WaitOne(this.Interval);
      } // while
      waitHandle.Close();
      abortThreadPending = false;
      Logger.Debug("Heartbeat thread stopped");
    }

    void wcfService_ProcessHeartBeatCompleted(object sender, ProcessHeartBeatCompletedEventArgs e) {
      Logger.Debug("Heartbeat Response received");
      e.Result.ActionRequest.ForEach(mc => MessageQueue.GetInstance().AddMessage(mc));
    }

    #region Eventhandler
    public event EventHandler<EventArgs<Exception>> ExceptionOccured;
    private void OnExceptionOccured(Exception e) {
      var handler = ExceptionOccured;
      if (handler != null) handler(this, new EventArgs<Exception>(e));
    }
    #endregion

    #region Helpers
    private int GetFreeMemory() {
      PerformanceCounter counter = new PerformanceCounter("Memory", "Available Bytes", true);
      int mb = (int)(counter.NextValue() / 1024 / 1024);
      return mb;
    }
    #endregion
  }
}
